import hashlib
from datetime import datetime, timedelta
import re
from typing import Optional

import requests
import xmltodict

from iikoserver_api.schemas.account import Account
from iikoserver_api.schemas.cashshift import Cashshift
from iikoserver_api.schemas.corporation import CorporationStore
from iikoserver_api.schemas.employee import Employee
from iikoserver_api.schemas.incoming_invoice import IncomingInvoice
from iikoserver_api.schemas.nomenclature_group import NomenclatureGroup
from iikoserver_api.schemas.outgoing_invoice import OutgoingInvoice
from iikoserver_api.schemas.product import Product
from iikoserver_api.schemas.write_off import WriteOff


class IikoServerApi:
    # _default_headers = {'Content-Type': 'application/xml'}

    def __init__(self, url, login, password):
        self.url = url
        self.login = login
        self.password = hashlib.sha1(password.encode()).hexdigest()
        self.token = None
        self.connected = False

    def send(self,
             action: str,
             method: str = 'GET',
             params: dict = None,
             headers: dict = None,
             data: any = None,
             json=None):
        # if headers is None:
        #     headers = self._default_headers
        url = f'{self.url}{action}/'
        response = requests.request(method=method, url=url, params=params, headers=headers, data=data, json=json)

        if response.ok:
            if action.startswith('v2'):
                return response.json()
            return response.content
        else:
            error_info = {
                "response_status": response.status_code,
                "response_info": response.content.decode()
            }
            raise ConnectionError(error_info)

    def auth(self):
        params = {
            'login': self.login,
            'pass': self.password,
        }
        response = self.send(action='auth', params=params)
        data = response.decode()
        self.token = data
        self.connected = True

    def logout(self):
        params = {
            'key': self.token
        }
        data = self.send(action='logout', params=params)
        self.token = None
        self.connected = False
        return data

    def export_outgoing_invoice(self, from_datetime: datetime,
                                to_datetime: datetime) -> Optional[list[OutgoingInvoice]]:
        params = {
            'key': self.token,
            'from': from_datetime.strftime('%Y-%m-%d'),
            'to': to_datetime.strftime('%Y-%m-%d'),
        }
        data = self.send(action='documents/export/outgoingInvoice', params=params)
        xml_dict_data = xmltodict.parse(data.decode())
        documents = xml_dict_data['outgoingInvoiceDtoes']
        if documents:
            documents = [documents['document']] if type(documents['document']) is dict else documents['document']
        else:
            return
        result = []
        for document in documents:
            document_items = document['items']['item']
            if document_items:
                document_items = [document['items']['item']] \
                    if type(document['items']['item']) is dict else document['items']['item']
            else:
                continue
            dict_data = {'items': []}
            for key, value in document.items():
                if key == 'items':
                    continue
                key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                dict_data[key] = value

            for item in document_items:
                new_item = {}
                for key, value in item.items():
                    key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                    new_item[key] = value

                dict_data['items'].append(new_item)

            result.append(OutgoingInvoice(**dict_data))

        return result

    def export_incoming_invoice(self, from_datetime: datetime,
                                to_datetime: datetime) -> Optional[list[IncomingInvoice]]:
        params = {
            'key': self.token,
            'from': from_datetime.strftime('%Y-%m-%d'),
            'to': to_datetime.strftime('%Y-%m-%d'),
        }
        data = self.send(action='documents/export/incomingInvoice', params=params)
        xml_dict_data = xmltodict.parse(data.decode())
        documents = xml_dict_data['incomingInvoiceDtoes']
        if documents:
            documents = [documents['document']] if type(documents['document']) is dict else documents['document']
        else:
            return
        result = []
        for document in documents:
            document_items = document['items']['item']
            if document_items:
                document_items = [document['items']['item']] \
                    if type(document['items']['item']) is dict else document['items']['item']
            else:
                continue
            dict_data = {'items': []}
            for key, value in document.items():
                if key == 'items':
                    continue
                key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                dict_data[key] = value

            for item in document_items:
                new_item = {}
                for key, value in item.items():
                    key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                    new_item[key] = value

                dict_data['items'].append(new_item)

            result.append(IncomingInvoice(**dict_data))

        return result

    def get_employee(self, pk: str) -> Optional[Employee]:
        params = {
            'key': self.token,
            # 'from': from_datetime.strftime('%Y-%m-%d'),
            # 'to': to_datetime.strftime('%Y-%m-%d'),
        }
        data = self.send(action=f'employees/byId/{pk}', params=params)
        xml_dict_data = xmltodict.parse(data.decode())
        employee = xml_dict_data['employee']
        dict_data = {}
        for key, value in employee.items():
            key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
            if value in ['true', 'false']:
                value = True if value == 'true' else False
            dict_data[key] = value

        return Employee(**dict_data)
        # documents = xml_dict_data['incomingInvoiceDtoes']

    def export_nomenclature_group(self):
        params = {
            'key': self.token,
        }
        data = self.send(action='v2/entities/products/group/list', params=params)
        result = []
        for el in data:
            result.append(NomenclatureGroup(**el))
        return result

    def export_products(self):
        params = {
            'key': self.token,
        }
        data = self.send(action='v2/entities/products/list', params=params)
        result = []
        for el in data:
            result.append(Product(**el))
        return result

    def export_cashshifts(self, from_datetime: datetime, to_datetime: datetime, status: str = 'ANY'):
        params = {
            'key': self.token,
            'openDateFrom': from_datetime.strftime('%Y-%m-%d'),
            'openDateTo': to_datetime.strftime('%Y-%m-%d'),
            'status': status,
        }
        data = self.send(action='v2/cashshifts/list', params=params)
        result = []
        for el in data:
            result.append(Cashshift(**el))
        return result

    def export_cashshift_payments(self, pk: str):
        params = {
            'key': self.token,
            'hideAccepted': False
        }
        data = self.send(action=f'v2/cashshifts/payments/list/{pk}', params=params)
        return data

    def export_stores(self) -> list[CorporationStore]:
        params = {
            'key': self.token,
        }
        data = self.send(action='corporation/stores', params=params)
        xml_dict_data = xmltodict.parse(data.decode())
        stores = xml_dict_data['corporateItemDtoes']['corporateItemDto']
        results = []
        for store in stores:
            dict_data = {}
            for key, value in store.items():
                key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                if value in ['true', 'false']:
                    value = True if value == 'true' else False
                dict_data[key] = value
            results.append(CorporationStore(**dict_data))

        return results

    def get_terminals(self):
        params = {
            'key': self.token,
        }
        data = self.send('corporation/departments', params=params)
        xml_dict_data = xmltodict.parse(data.decode())
        return xml_dict_data

    def export_pal(self):
        params = {
            'key': self.token,
            'reportType': 'TRANSACTIONS'
        }
        end_date = datetime.now()
        start_date = end_date - timedelta(days=3)
        data = {
            "reportType": "TRANSACTIONS",
            "buildSummary": "false",
            "groupByRowFields": [],
            "groupByColFields": [
                "Account.Name",
                "Account.Id",
                "Account.StoreOrAccount",
            ],
            "filters": {
                "DateTime.DateTyped": {
                    "filterType": "DateRange",
                    "periodType": "CUSTOM",
                    "from": start_date.strftime('%Y-%m-%d'),
                    "to": end_date.strftime('%Y-%m-%d'),
                    "includeLow": True,
                    "includeHigh": True
                }
            }
        }
        data = self.send('v2/reports/olap', method='POST', json=data, params=params)
        return data

    def export_write_offs(self, from_datetime: datetime, to_datetime: datetime,):
        params = {
            'key': self.token,
            'dateFrom': from_datetime.strftime('%Y-%m-%d'),
            'dateTo': to_datetime.strftime('%Y-%m-%d'),
        }
        data = self.send(action=f'v2/documents/writeoff', params=params)
        result = []
        for el in data.get('response', []):
            result.append(WriteOff(**el))
        return result

    def export_accounts(self):
        params = {
            'key': self.token,
            'reportType': 'TRANSACTIONS'
        }
        end_date = datetime.now()
        start_date = end_date - timedelta(days=90)
        json = {
            "reportType": "TRANSACTIONS",
            "buildSummary": "false",
            "groupByRowFields": [],
            "groupByColFields": [
                "Account.Name",
                "Account.Id",
                "Account.StoreOrAccount",
            ],
            "filters": {
                "DateTime.DateTyped": {
                    "filterType": "DateRange",
                    "periodType": "CUSTOM",
                    "from": start_date.strftime('%Y-%m-%d'),
                    "to": end_date.strftime('%Y-%m-%d'),
                    "includeLow": True,
                    "includeHigh": True
                },
                "Account.StoreOrAccount": {
                    "filterType": "IncludeValues",
                    "values": ["ACCOUNT"]
                }
            }
        }
        data = self.send('v2/reports/olap', method='POST', json=json, params=params)
        result = []
        for item in data.get('data', []):
            result.append(Account(**item))
        return result

    def export_suppliers(self) -> list[Employee]:
        params = {
            'key': self.token,
        }
        data = self.send(action='suppliers', params=params)
        xml_dict_data = xmltodict.parse(data.decode())
        employees = xml_dict_data['employees']['employee']
        result = []
        for employee in employees:
            dict_data = {}
            for key, value in employee.items():
                key = re.sub(r'(?<!^)(?=[A-Z])', '_', key).lower()
                if value in ['true', 'false']:
                    value = True if value == 'true' else False
                dict_data[key] = value
            result.append(Employee(**dict_data))

        return result

    def __enter__(self):
        self.auth()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logout()


if __name__ == '__main__':
    with IikoServerApi('https://tutto.iiko.it/resto/api/', 'rsinfo', '5E6MYvS2KO') as api:
        data = api.export_suppliers()
        print(data)
